from .wrapper import *
from .utils import *
from .matrix import *
from .io import *

# Setup global module state
init_wrapper()

print("**pyspbla project**")
print("pyspbla is a python wrapper for a spbla library.")
print("See https://github.com/JetBrains-Research/spbla for more information.")
